def url_pagination(params: dict) -> None:
    if "page" not in params or "per_page" not in params:
        raise ValueError("'page' and 'per_page' parameters are required to requests" " with pagination")



import asyncio
import concurrent.futures
import time
from itertools import chain
from typing import AnyStr, Generator, Literal
from urllib.parse import urljoin

import aiohttp
import requests

from mosqlient.config import API_DEV_URL, API_PROD_URL, APPS, APPS_TYPE
from mosqlient.utils import validate


def get(
    app: APPS_TYPE,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    pagination: bool = True,
    timeout: int = 10,
    _env: Literal["dev", "prod"] = "prod",
) -> requests.models.Response:
    if app not in APPS:
        raise ValueError(f"unkown Mosqlimate app. Options: {APPS}")

    if pagination:
        validate.url_pagination(params)

    if not endpoint:
        raise ValueError("endpoint is required")

    base_url = API_DEV_URL if _env == "dev" else API_PROD_URL

    url = urljoin(base_url, "/".join((str(app), str(endpoint)))) + "/?"

    return requests.get(url, params, timeout=timeout)


async def aget(
    session: aiohttp.ClientSession, url: str, params: dict[str, str | int | float], timeout: int = 10, retries: int = 3
) -> dict:
    try:
        if retries < 0:
            raise aiohttp.ClientConnectionError("Too many attempts")
        async with session.get(url, params=params, timeout=timeout) as res:
            if res.status == 200:
                return await res.json()
            if retries == 0:
                raise aiohttp.ClientConnectionError(f"Response status: {res.status}. Reason: {res.reason}")
            await asyncio.sleep(10 / (retries + 1))
            await aget(session, url, params, timeout, retries - 1)
    except aiohttp.ServerTimeoutError:
        await asyncio.sleep(8 / (retries + 1))
        await aget(session, url, params, timeout, retries - 1)
    raise aiohttp.ClientConnectionError("Invalid request")


async def get_all(
    app: APPS_TYPE,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    timeout: int = 60,
    _max_per_page: int = 50,
    _env: Literal["dev", "prod"] = "prod",
) -> list[dict]:
    params["page"] = 1
    params["per_page"] = _max_per_page

    base_url = API_DEV_URL if _env == "dev" else API_PROD_URL

    url = urljoin(base_url, "/".join((str(app), str(endpoint)))) + "/?"

    async with aiohttp.ClientSession() as session:
        first_page = await aget(session, url, params)

    total_pages = first_page["pagination"]["total_pages"]

    if total_pages == 1:
        return first_page["items"]

    async with aiohttp.ClientSession(timeout=timeout) as session:
        tasks = []
        for page in range(2, total_pages + 1):
            params["page"] = page
            task = asyncio.create_task(aget(session, url, params))
            tasks.append(task)
        results = await asyncio.gather(*tasks)
    return list(chain(result["items"] for result in results))


def compose_url(base_url: str, parameters: dict, page: int = 1) -> str:
    """Helper method to compose the API url with parameters"""
    url = base_url + "?" if not base_url.endswith("?") else base_url
    params = "&".join([f"{p}={v}" for p, v in parameters.items()]) + f"&page={page}"
    return url + params


def fetch_data(session: requests.Session, url: str):
    """Uses ClientSession to create the async call to the API"""
    response = session.get(url)
    return response.json()


def attempt_delay(session: requests.Session, url: str):
    """The request may fail. This method adds a delay to the failing requests"""
    try:
        return fetch_data(session, url)
    except Exception as e:
        time.sleep(0.2)
        return attempt_delay(session, url)


def get_datastore(
    app: APPS_TYPE, endpoint: AnyStr, params: dict[str, str | int | float], _env: Literal["dev", "prod"] = "prod"
) -> Generator[dict, None, None]:

    base_url = API_DEV_URL if _env == "dev" else API_PROD_URL

    base_url = urljoin(base_url, "/".join((str(app), str(endpoint)))) + "/?"

    result = []
    with requests.Session() as session:
        url = compose_url(base_url, params)
        data = attempt_delay(session, url)
        total_pages = data["pagination"]["total_pages"]
        result.extend(data["items"])
        if total_pages == 1:
            for i in data["items"]:
                yield i

        futures = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # Start the load operations and mark each future with its URL
            for page in range(2, total_pages + 1):
                url = compose_url(base_url, params, page)
                futures[executor.submit(attempt_delay, session, url)] = url
            for future in concurrent.futures.as_completed(futures):
                _ = futures[future]
                resp = future.result()
                if result:  # incorporating the first page
                    resp["items"].extend(result)
                    result = []
                for i in resp["items"]:
                    yield i



import uuid

from mosqlient.requests import get


class PlatformClient:
    def __init__(self, x_uid_key: str):
        self.username, self.uid_key = x_uid_key.split(":")
        self._check_username()
        self._check_uuid4()

    def __str__(self):
        return self.username

    @property
    def X_UID_KEY(self):
        return f"{self.username}:{self.uid_key}"

    def _check_username(self):
        author = get("registry", "authors", {"username": self.username}, pagination=False)

        if author.status_code != 200:
            raise ValueError(f"Could not get user '{self.username}' info. " f"Status code: {author.status_code}")

    def _check_uuid4(self):
        try:
            uuid.UUID(self.uid_key, version=4)
        except ValueError:
            raise ValueError("uid_key is not a valid key")



from datetime import datetime
from typing import Any

import pandas as pd

from mosqlient.requests import get_datastore


def validate_date(date_str: str, par_name: str) -> None:
    try:
        # Validate the date format
        valid_date = datetime.strptime(date_str, "%Y-%m-%d")
    except ValueError:
        raise ValueError(f"Invalid date format for '{par_name}', should be YYYY-mm-dd")


def _params(**kwargs) -> dict[str, Any]:
    params = {}
    for k, v in kwargs.items():
        if isinstance(v, (bool, int, float, str)):
            params[k] = str(v)
        elif v is None:
            continue
        else:
            raise TypeError(f"Unknown type f{type(v)}")

    return params


class Infodengue:
    @classmethod
    def get(cls, **kwargs):
        cls._validate_fields(**kwargs)
        params = _params(**kwargs)

        gen_data = get_datastore("datastore", "infodengue", params)

        return pd.DataFrame(list(gen_data))

    @staticmethod
    def _validate_fields(**kwargs) -> None:
        DataFieldValidator(**kwargs)


class Copernicus:
    @classmethod
    def get(cls, **kwargs):
        cls._validate_fields(**kwargs)
        params = _params(**kwargs)

        gen_data = get_datastore("datastore", "climate", params)

        return pd.DataFrame(list(gen_data))

    @staticmethod
    def _validate_fields(**kwargs) -> None:
        DataFieldValidator(**kwargs)


class DataFieldValidator:
    FIELDS = {
        "per_page": int,
        "disease": str,
        "start": str,
        "end": str,
        "uf": str,
        "geocode": int,
    }
    DISEASES = ["dengue", "zika", "chikungunya"]

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if v is None:
                continue

            if not isinstance(v, self.FIELDS[k]):
                raise TypeError(f"Field '{k}' must have instance of " f"{' or '.join(self.FIELDS[k])}")

            if k == "disease":
                if v == "chik":
                    v = "chikungunya"

                if v not in self.DISEASES:
                    raise ValueError(f"Unkown 'disease'. Options: {self.DISEASES}")

            if (k == "start") | (k == "end"):
                validate_date(v, k)

            if k == "uf":
                if len(v) != 2:
                    raise ValueError("Invalid 'uf' parameter. It should be a two-letter value")

            if k == "geocode":
                if len(str(v)) != 7:
                    raise ValueError("Invalid 'geocode' parameter. It should be a seven numbers code")



import typing

APPS = ["registry", "datastore", "vis"]
APPS_TYPE = typing.Literal["registry", "datastore", "vis"]
API_DEV_URL = "http://0.0.0.0:8042/api/"
API_PROD_URL = "https://api.mosqlimate.org/api/"
